@@ -1,14 +1,13 @@
 """Assorted utilities used by drivers and the CLI.
 
-Copyright (C) 2018–2021  Jonas Malaco and contributors
+Copyright (C) 2018–2022  Jonas Malaco and contributors
 SPDX-License-Identifier: GPL-3.0-or-later
 """
 
 import colorsys
 import logging
-
 from ast import literal_eval
-from enum import Enum, unique
+from enum import Enum, EnumMeta, unique
 
 from liquidctl.error import UnsafeFeaturesNotEnabled
 
@@ -97,6 +96,16 @@ class LazyHexRepr:
         return self.sep.join(hexvals)
 
 
+class _RelaxedNamesEnum(EnumMeta):
+    def __getitem__(cls, name):
+        return super().__getitem__(name.upper())
+
+
+class RelaxedNamesEnum(Enum, metaclass=_RelaxedNamesEnum):
+    """Enum class where name lookup is case insensitive."""
+    pass
+
+
 def rpadlist(list, width, fillitem=0):
     """Pad `list` with `fillitem` to `width`.
 
@@ -360,3 +369,33 @@ def check_unsafe(*reqs, unsafe=None, error=False, **kwargs):
     if error:
         raise UnsafeFeaturesNotEnabled(*reqs)
     return False
+
+
+def map_direction(direction, forward=None, backward=None):
+    """Check a `direction` option and run the appropriate closure.
+
+    Unstable.
+
+    Accepts both US and UK spellings.  Raises a `ValueError` if the option
+    value is neither forward[s] nor backward[s].
+
+    >>> map_direction('forward', 3, 42)
+    3
+    >>> map_direction('forwards', 3, 42)
+    3
+    >>> map_direction('backward', 3, 42)
+    42
+    >>> map_direction('backwards', 3, 42)
+    42
+    >>> map_direction('fooowwd', 3, 42)
+    Traceback (most recent call last):
+        ...
+    ValueError: invalid direction: 'fooowwd'
+    """
+
+    if 'forwards'.startswith(direction):
+        return forward
+    elif 'backwards'.startswith(direction):
+        return backward
+    else:
+        raise ValueError(f'invalid direction: {direction!r}')
